@@ -22,7 +22,7 @@ __all__ = [
     'DecimalField', 'CharField', 'TextField', 'DateTimeField', 'DateField', 'TimeField',
     'BooleanField', 'ForeignKeyField', 'Model', 'DoesNotExist', 'ImproperlyConfigured',
     'DQ', 'fn', 'SqliteDatabase', 'MySQLDatabase', 'PostgresqlDatabase', 'Field',
-    'JOIN_LEFT_OUTER', 'JOIN_INNER', 'JOIN_FULL',
+    'JOIN_LEFT_OUTER', 'JOIN_INNER', 'JOIN_FULL', 'prefetch',
 ]
 
 try:
@@ -108,19 +108,43 @@ def dict_update(orig, extra):
     new.update(extra)
     return new
 
+def returns_clone(func):
+    def inner(self, *args, **kwargs):
+        clone = self.clone()
+        func(clone, *args, **kwargs)
+        return clone
+    inner.call_local = func
+    return inner
+
+def not_allowed(fn):
+    def inner(self, *args, **kwargs):
+        raise NotImplementedError('%s is not allowed on %s instances' % (
+            fn, type(self).__name__,
+        ))
+    return inner
+
 
 class Leaf(object):
     def __init__(self):
         self.negated = False
         self._alias = None
 
+    def clone_base(self):
+        return type(self)()
+
+    def clone(self):
+        inst = self.clone_base()
+        inst.negated = self.negated
+        inst._alias = self._alias
+        return inst
+
+    @returns_clone
     def __invert__(self):
         self.negated = not self.negated
-        return self
 
-    def alias(self, a):
+    @returns_clone
+    def alias(self, a=None):
         self._alias = a
-        return self
 
     def asc(self):
         return Ordering(self, True)
@@ -163,22 +187,21 @@ class Leaf(object):
 
 
 class Expr(Leaf):
-    def __init__(self, lhs, op, rhs, negated=False):
+    def __init__(self, lhs, op, rhs):
         super(Expr, self).__init__()
         self.lhs = lhs
         self.op = op
         self.rhs = rhs
-        self.negated = negated
 
-    def clone(self):
-        return Expr(self.lhs, self.op, self.rhs, self.negated)
+    def clone_base(self):
+        return Expr(self.lhs, self.op, self.rhs)
 
 class DQ(Leaf):
     def __init__(self, **query):
         super(DQ, self).__init__()
         self.query = query
 
-    def clone(self):
+    def clone_base(self):
         return DQ(**self.query)
 
 class Param(Leaf):
@@ -186,24 +209,33 @@ class Param(Leaf):
         self.data = data
         super(Param, self).__init__()
 
+    def clone_base(self):
+        return Param(self.data)
+
 class R(Leaf):
     def __init__(self, value):
         self.value = value
         super(R, self).__init__()
 
+    def clone_base(self):
+        return R(self.value)
+
 class Ordering(Leaf):
     def __init__(self, param, asc):
         self.param = param
         self.asc = asc
         super(Ordering, self).__init__()
 
+    def clone_base(self):
+        return Ordering(self.param, self.asc)
+
 class Func(Leaf):
     def __init__(self, name, *params):
         self.name = name
         self.params = params
         super(Func, self).__init__()
 
-    def clone(self):
+    def clone_base(self):
         return Func(self.name, *self.params)
 
     def __getattr__(self, attr):
@@ -256,6 +288,25 @@ class Field(Leaf):
 
         super(Field, self).__init__()
 
+    def clone_base(self, **kwargs):
+       inst = type(self)(
+           null=self.null,
+           index=self.index,
+           unique=self.unique,
+           help_text=self.help_text,
+           default=self.default,
+           choices=self.choices,
+           primary_key=self.primary_key,
+           sequence=self.sequence,
+           **kwargs
+       )
+       inst.attributes = dict(self.attributes)
+       inst.name = self.name
+       inst.model_class = self.model_class
+       inst.db_column = self.db_column
+       inst.verbose_name = self.verbose_name
+       return inst
+
     def add_to_class(self, model_class, name):
         self.name = name
         self.model_class = model_class
@@ -490,6 +541,16 @@ class ForeignKeyField(IntegerField):
 
         super(ForeignKeyField, self).__init__(null=null, *args, **kwargs)
 
+    def clone_base(self):
+        inst = super(ForeignKeyField, self).clone_base(
+            rel_model=self.rel_model,
+            cascade=self.cascade,
+            extra=self.extra,
+        )
+        inst.related_name = self.related_name
+        inst.rel_model = self.rel_model
+        return inst
+
     def add_to_class(self, model_class, name):
         self.name = name
         self.model_class = model_class
@@ -702,40 +763,43 @@ class QueryCompiler(object):
 
     def generate_joins(self, joins, model_class, alias_map):
         parsed = []
+        params = []
         seen = set()
-
-        def _traverse(curr):
+        q = [model_class]
+        while q:
+            curr = q.pop()
             if curr not in joins or curr in seen:
-                return
+                continue
             seen.add(curr)
             for join in joins[curr]:
                 from_model = curr
                 to_model = join.model_class
-
-                field = from_model._meta.rel_for_model(to_model, join.on)
-                if field:
-                    left_field = field.db_column
-                    right_field = to_model._meta.primary_key.db_column
+                if isinstance(join.on, Expr):
+                    join_expr = join.on
                 else:
-                    field = to_model._meta.rel_for_model(from_model, join.on)
-                    left_field = from_model._meta.primary_key.db_column
-                    right_field = field.db_column
+                    field = from_model._meta.rel_for_model(to_model, join.on)
+                    if field:
+                        left_field = field
+                        right_field = to_model._meta.primary_key
+                    else:
+                        field = to_model._meta.rel_for_model(from_model, join.on)
+                        left_field = from_model._meta.primary_key
+                        right_field = field
+                    join_expr = (left_field == right_field)
 
                 join_type = join.join_type or JOIN_INNER
-                lhs = '%s.%s' % (alias_map[from_model], self.quote(left_field))
-                rhs = '%s.%s' % (alias_map[to_model], self.quote(right_field))
+                parsed_join, join_params = self.parse_expr(join_expr, alias_map)
 
-                parsed.append('%s JOIN %s AS %s ON %s = %s' % (
+                parsed.append('%s JOIN %s AS %s ON %s' % (
                     self.join_map[join_type],
                     self.quote(to_model._meta.db_table),
                     alias_map[to_model],
-                    lhs,
-                    rhs,
+                    parsed_join,
                 ))
+                params.extend(join_params)
 
-                _traverse(to_model)
-        _traverse(model_class)
-        return parsed
+                q.append(to_model)
+        return parsed, params
 
     def generate_select(self, query, start=1, alias_map=None):
         model = query.model_class
@@ -758,9 +822,10 @@ class QueryCompiler(object):
 
         parts.append('FROM %s AS %s' % (self.quote(model._meta.db_table), alias_map[model]))
 
-        joins = self.generate_joins(query._joins, query.model_class, alias_map)
+        joins, j_params = self.generate_joins(query._joins, query.model_class, alias_map)
         if joins:
             parts.append(' '.join(joins))
+            params.extend(j_params)
 
         where, w_params = self.parse_query_node(query._where, alias_map)
         if where:
@@ -942,14 +1007,17 @@ class QueryResultWrapper(object):
         cols = [c[0] for c in self.cursor.description]
         for i, expr in enumerate(self.column_meta):
             value = row[i]
-            if isinstance(expr, Field):
-                model = expr.model_class
+            if isinstance(expr, FieldProxy):
+                key = expr._model_alias # model alias
+                constructor = expr.model # instance constructor
+            elif isinstance(expr, Field):
+                key = constructor = expr.model_class
             else:
-                model = self.model
+                key = constructor = self.model
 
-            if model not in collected_models:
-                collected_models[model] = model()
-            instance = collected_models[model]
+            if key not in collected_models:
+                collected_models[key] = constructor()
+            instance = collected_models[key]
 
             if isinstance(expr, Field):
                 setattr(instance, expr.name, expr.python_value(value))
@@ -958,29 +1026,35 @@ class QueryResultWrapper(object):
             else:
                 setattr(instance, cols[i], value)
 
-        return self.follow_joins(self.join_meta, collected_models, self.model)
+        return collected_models
 
-    def follow_joins(self, joins, collected_models, current):
-        inst = collected_models[current]
+    def follow_joins(self, collected):
+        joins = self.join_meta
+        stack = [self.model]
+        while stack:
+            current = stack.pop()
+            if current not in joins:
+                continue
 
-        if current not in joins:
-            return inst
+            inst = collected[current]
+            for join in joins[current]:
+                if join.model_class in collected:
+                    joined_inst = collected[join.model_class]
+                    fk_field = current._meta.rel_for_model(join.model_class)
+                    if not fk_field:
+                        if isinstance(join.on, Expr):
+                            fk_field = join.on.lhs
+                        else:
+                            continue
 
-        for joined_model, _, _ in joins[current]:
-            if joined_model in collected_models:
-                joined_inst = self.follow_joins(joins, collected_models, joined_model)
-                fk_field = current._meta.rel_for_model(joined_model)
+                    if joined_inst.get_id() is None and fk_field.name in inst._data:
+                        rel_inst_id = inst._data[fk_field.name]
+                        joined_inst.set_id(rel_inst_id)
 
-                if not fk_field:
-                    continue
+                    setattr(inst, fk_field.name, joined_inst)
+                    stack.append(join.model_class)
 
-                if joined_inst.get_id() is None and fk_field.name in inst._data:
-                    rel_inst_id = inst._data[fk_field.name]
-                    joined_inst.set_id(rel_inst_id)
-
-                setattr(inst, fk_field.name, joined_inst)
-
-        return inst
+        return collected[self.model]
 
     def __iter__(self):
         self.__idx = 0
@@ -999,7 +1073,8 @@ class QueryResultWrapper(object):
         if self.naive:
             return self.simple_iter(row)
         else:
-            return self.construct_instance(row)
+            collected = self.construct_instance(row)
+            return self.follow_joins(collected)
 
     def iterator(self):
         while 1:
@@ -1027,28 +1102,14 @@ class QueryResultWrapper(object):
             except StopIteration:
                 break
 
-def returns_clone(func):
-    def inner(self, *args, **kwargs):
-        clone = self.clone()
-        func(clone, *args, **kwargs)
-        return clone
-    inner.call_local = func
-    return inner
-
-def not_allowed(fn):
-    def inner(self, *args, **kwargs):
-        raise NotImplementedError('%s is not allowed on %s instances' % (
-            fn, type(self).__name__,
-        ))
-    return inner
-
-
 Join = namedtuple('Join', ('model_class', 'join_type', 'on'))
 
-class Query(object):
+class Query(Leaf):
     require_commit = True
 
     def __init__(self, model_class):
+        super(Query, self).__init__()
+
         self.model_class = model_class
         self.database = model_class._meta.database
 
@@ -1084,7 +1145,7 @@ class Query(object):
 
     @returns_clone
     def join(self, model_class, join_type=None, on=None):
-        if not self._query_ctx._meta.rel_exists(model_class):
+        if not self._query_ctx._meta.rel_exists(model_class) and on is None:
             raise ValueError('No foreign key between %s and %s' % (
                 self._query_ctx, model_class,
             ))
@@ -1221,6 +1282,7 @@ class SelectQuery(Query):
         self._distinct = False
         self._for_update = False
         self._naive = False
+        self._alias = None
         self._qr = None
 
     def clone(self):
@@ -1238,6 +1300,7 @@ class SelectQuery(Query):
         query._distinct = self._distinct
         query._for_update = self._for_update
         query._naive = self._naive
+        query._alias = self._alias
         return query
 
     def _model_shorthand(self, args):
@@ -1245,6 +1308,10 @@ class SelectQuery(Query):
         for arg in args:
             if isinstance(arg, Leaf):
                 accum.append(arg)
+            elif isinstance(arg, Query):
+                accum.append(arg)
+            elif isinstance(arg, ModelAlias):
+                accum.extend(arg.get_proxy_fields())
             elif issubclass(arg, Model):
                 accum.extend(arg._meta.get_fields())
         return accum
@@ -1292,12 +1359,16 @@ class SelectQuery(Query):
     def naive(self, naive=True):
         self._naive = naive
 
+    @returns_clone
+    def alias(self, alias=None):
+        self._alias = alias
+
     def annotate(self, rel_model, annotation=None):
         annotation = annotation or fn.Count(rel_model._meta.primary_key).alias('count')
         query = self.clone()
         query = query.ensure_join(query._query_ctx, rel_model)
         if not query._group_by:
-            query._group_by = list(query._select)
+            query._group_by = [x.alias() for x in query._select]
         query._select = tuple(query._select) + (annotation,)
         return query
 
@@ -1339,7 +1410,7 @@ class SelectQuery(Query):
             raise self.model_class.DoesNotExist('instance matching query does not exist:\nSQL: %s\nPARAMS: %s' % (
                 self.sql()
             ))
-    
+
     def first(self):
         res = self.execute()
         res.fill_cache(1)
@@ -1767,11 +1838,15 @@ class transaction(object):
         self.db.begin()
 
     def __exit__(self, exc_type, exc_val, exc_tb):
+        success = True
         if exc_type:
             self.db.rollback()
+            success = False
         else:
             self.db.commit()
         self.db.set_autocommit(self._orig)
+        return success
+
 
 class DoesNotExist(Exception):
     pass
@@ -1824,7 +1899,7 @@ class ModelOptions(object):
         return dd
 
     def get_sorted_fields(self):
-        return sorted(self.fields.items(), key=lambda (k,v): (v is self.primary_key and 1 or 2, v._order))
+        return sorted(self.fields.items(), key=lambda i: (i[1] is self.primary_key and 1 or 2, i[1]._order))
 
     def get_field_names(self):
         return [f[0] for f in self.get_sorted_fields()]
@@ -1910,6 +1985,37 @@ class BaseModel(type):
         return cls
 
 
+class FieldProxy(Field):
+    def __init__(self, alias, field_instance):
+        self._model_alias = alias
+        self.model = self._model_alias.model_class
+        self.field_instance = field_instance
+
+    def clone_base(self):
+        return FieldProxy(self._model_alias, self.field_instance)
+
+    def __getattr__(self, attr):
+        if attr == 'model_class':
+            return self._model_alias
+        return getattr(self.field_instance, attr)
+
+class ModelAlias(object):
+    def __init__(self, model_class):
+        self.__dict__['model_class'] = model_class
+
+    def __getattr__(self, attr):
+        model_attr = getattr(self.model_class, attr)
+        if isinstance(model_attr, Field):
+            return FieldProxy(self, model_attr)
+        return model_attr
+
+    def __setattr__(self, attr, value):
+        raise AttributeError('Cannot set attributes on ModelAlias instances')
+
+    def get_proxy_fields(self):
+        return [FieldProxy(self, f) for f in self.model_class._meta.get_fields()]
+
+
 class Model(object):
     __metaclass__ = BaseModel
 
@@ -1920,6 +2026,10 @@ class Model(object):
         for k, v in kwargs.items():
             setattr(self, k, v)
 
+    @classmethod
+    def alias(cls):
+        return ModelAlias(cls)
+
     @classmethod
     def select(cls, *selection):
         query = SelectQuery(cls, *selection)
@@ -2011,11 +2121,20 @@ class Model(object):
     def prepared(self):
         pass
 
-    def save(self, force_insert=False):
+    def _prune_fields(self, field_dict, only):
+        new_data = {}
+        for field in only:
+            if field.name in field_dict:
+                new_data[field.name] = field_dict[field.name]
+        return new_data
+
+    def save(self, force_insert=False, only=None):
         field_dict = dict(self._data)
         pk = self._meta.primary_key
+        if only:
+            field_dict = self._prune_fields(field_dict, only)
         if self.get_id() is not None and not force_insert:
-            field_dict.pop(pk.name)
+            field_dict.pop(pk.name, None)
             update = self.update(
                 **field_dict
             ).where(pk == self.get_id())
@@ -2062,6 +2181,56 @@ class Model(object):
         return not self == other
 
 
+def prefetch_add_subquery(sq, subqueries):
+    fixed_queries = [(sq, None)]
+    for i, subquery in enumerate(subqueries):
+        if not isinstance(subquery, Query) and issubclass(subquery, Model):
+            subquery = subquery.select()
+        subquery_model = subquery.model_class
+        fkf = None
+        for j in range(i + 1):
+            last_query = fixed_queries[i - j - 1][0]
+            fkf = subquery_model._meta.rel_for_model(last_query.model_class)
+            if fkf:
+                break
+        if not fkf:
+            raise AttributeError('Error: unable to find foreign key for query: %s' % subquery)
+        fixed_queries.append((subquery.where(fkf << last_query), fkf))
+
+    return fixed_queries
+
+def prefetch(sq, *subqueries):
+    if not subqueries:
+        return sq
+    fixed_queries = prefetch_add_subquery(sq, subqueries)
+
+    deps = {}
+    rel_map = {}
+    for query, foreign_key_field in reversed(fixed_queries):
+        query_model = query.model_class
+        deps[query_model] = {}
+        id_map = deps[query_model]
+        has_relations = bool(rel_map.get(query_model))
+
+        for result in query:
+            if foreign_key_field:
+                fk_val = result._data[foreign_key_field.name]
+                id_map.setdefault(fk_val, [])
+                id_map[fk_val].append(result)
+            if has_relations:
+                for rel_model, rel_fk in rel_map[query_model]:
+                    rel_name = '%s_prefetch' % rel_fk.related_name
+                    rel_instances = deps[rel_model].get(result.get_id(), [])
+                    for inst in rel_instances:
+                        setattr(inst, rel_fk.name, result)
+                    setattr(result, rel_name, rel_instances)
+        if foreign_key_field:
+            rel_model = foreign_key_field.rel_model
+            rel_map.setdefault(rel_model, [])
+            rel_map[rel_model].append((query_model, foreign_key_field))
+
+    return query
+
 def create_model_tables(models, **create_table_kwargs):
     """Create tables for all given models (in the right order)."""
     for m in sort_models_topologically(models):
